/*
 *  Interactive disassembler (IDA).
 *  TheHyper VM Processor Module.
 *
 *  (c) 2006 Rolf Rolles
 */

#include "ins.hpp"
#include "TH.hpp"

// THINSTENC == TheHyper INSTruction ENCoding
// used generally
#define THINSTENC_LargeWidthModifier        0x1
#define THINSTENC_SmallWidthModifier        0x2
#define THINSTENC_4                         0x4
#define THINSTENC_DereferenceSecondOperand  0x8
#define THINSTENC_SecondOperandIsRegister  0x10
#define THINSTENC_DisplaceSecondOperand    0x20
#define THINSTENC_DereferenceFirstOperand  0x40
#define THINSTENC_FirstOperandIsRegister   0x80

// used for the jmp/call instructions
#define THINSTENC_DereferenceOperand       0x10
#define THINSTENC_HeapMemoryKeyFollows     0x20
#define THINSTENC_OneByteDisplacement      0x40 // otherwise, it's a dword

unsigned long WidthToDtyp(unsigned long width)
{
// #define dt_byte         0       // 8 bit
// #define dt_word         1       // 16 bit
// #define dt_dword        2       // 32 bit

    return width >> 1; 
}

unsigned long ana_th_0_to_0d(unsigned long code)
{
    bool          bRegister      = 0;
    TH_regs       reg1           = rEND;
    TH_regs       reg2           = rEND;
    unsigned long width          = 1; // bitness of registers, immediates
    unsigned char length         = 4;
    unsigned char encoding_opts  = ua_next_byte();
    unsigned char register_usage = ua_next_byte();
    unsigned char displacement   = ua_next_byte();

    switch(code)
    {
    case 0:
        cmd.itype = TH_mov;
        break;
    case 1:
        cmd.itype = TH_add;
        break;
    case 2:
        cmd.itype = TH_sub;
        break;
    case 3:
        cmd.itype = TH_xor;
        break;
    case 4:
        cmd.itype = TH_and;
        break;
    case 5:
        cmd.itype = TH_or;
        break;
    case 6:
        cmd.itype = TH_imul;
        break;
    case 7:
        cmd.itype = TH_idiv;
        break;
    case 8:
        cmd.itype = TH_imod;
        break;
    case 9:
        cmd.itype = TH_ror;
        break;
    case 10:
        cmd.itype = TH_rol;
        break;
    case 11:
        cmd.itype = TH_shr;
        break;
    case 12:
        cmd.itype = TH_shl;
        break;
    case 13:
        cmd.itype = TH_cmp;
        break;
    }
    
    if(encoding_opts & THINSTENC_LargeWidthModifier)
        width += 2;

    if( !(encoding_opts & THINSTENC_SmallWidthModifier) )
        width += 1;

    cmd.Op1.dtyp = WidthToDtyp(width);
    if(encoding_opts & THINSTENC_FirstOperandIsRegister)
    {
        bRegister = 1;
        reg1      = TH_Regnum(width, (register_usage & 0xf0) >> 4);

        // is this a memory dereference?
        // i.e. we go from o_reg to o_displ or o_phrase here
        if(encoding_opts & THINSTENC_DereferenceFirstOperand)
        {
            cmd.Op1.phrase = reg1;
            if(displacement != 0)
            {
                cmd.Op1.type   = o_displ;
                cmd.Op1.addr   = displacement;
            }
            else
                cmd.Op1.type   = o_phrase;
        }
        else
        {
            cmd.Op1.type = o_reg;
            cmd.Op1.reg  = reg1;
        }
    }
    else
    {
        // o_imm
        cmd.Op1.type  = o_imm;

        if(width == 1)
            cmd.Op1.value = ua_next_byte();
        else
        if(width == 2)
            cmd.Op1.value = ua_next_word();
        else
        if(width == 4)
            cmd.Op1.value = ua_next_long();

        length  += width;
    }
    bool bSecondRegister;
    unsigned long second_value = 0; 
    unsigned long second_displ = 0;
    cmd.Op2.dtyp = WidthToDtyp(width);

    if(encoding_opts & THINSTENC_SecondOperandIsRegister)
    {
        reg2            = TH_Regnum(width, register_usage & 0xf);
        bSecondRegister = 1;
    }

    else
    {
        bSecondRegister = 0;

        if(width == 1)
            second_value = ua_next_byte();
        else
        if(width == 2)
            second_value = ua_next_word();
        else
        if(width == 4)
            second_value = ua_next_long();

        length  += width;
    }

    if( encoding_opts & THINSTENC_DereferenceSecondOperand )
    {
        if( encoding_opts & THINSTENC_DisplaceSecondOperand )
        {
            if(!bSecondRegister)
                second_value += displacement;

        }
        // if first is register , and we do    dereference, we have o_displ
        if(bSecondRegister)
        {
            cmd.Op2.type   = o_phrase;
            cmd.Op2.phrase = reg2;
            if(displacement != 0)
            {
                cmd.Op2.addr   = displacement;
                cmd.Op2.type   = o_displ;
            }
        }
        // if first is immediate, and we do    dereference, we have o_mem
        else
        {
            cmd.Op2.type   = o_mem;
            cmd.Op2.addr   = second_value;
        }

    }
    else
    {
        // if first is register , and we don't dereference, we have o_reg
        if(bSecondRegister)
        {
            cmd.Op2.type = o_reg;
            cmd.Op2.reg  = reg2;
        }
        // if first is immediate, and we don't dereference, we have o_imm
        else
        {
            cmd.Op2.type  = o_imm;
            cmd.Op2.value = second_value;
        }
    }
    
    return length;

    // cl  = function # code
    // dx  = length (length of instr, added to EIP to determine next EIP)
    // ebx = width >> 1
    // esi = live
    // edi = live?
}

unsigned long ana_th_0e_to_12(unsigned long code)
{
    TH_regs       reg1           = rEND;
    unsigned long width          = 1; // bitness of registers, immediates
    unsigned char length         = 3;
    unsigned char encoding_opts  = ua_next_byte();

    switch(code)
    {
    case 14:
        cmd.itype = TH_push;
        break;
    case 15:
        cmd.itype = TH_pop;
        break;
    case 16:
        cmd.itype = TH_inc;
        break;
    case 17:
        cmd.itype = TH_dec;
        break;
    case 18:
        cmd.itype = TH_not;
        break;
    }

    if(encoding_opts & THINSTENC_LargeWidthModifier)
        width += 2;

    if( !(encoding_opts & THINSTENC_SmallWidthModifier) )
        width += 1;

    cmd.Op1.dtyp = WidthToDtyp(width);

    // are we manipulating the contents of a register?
    if(encoding_opts & THINSTENC_FirstOperandIsRegister)    
    {
        reg1 = TH_Regnum(width, (encoding_opts >> 2) & 0xf);
        
        // do we have a displaced phrase?
        if(encoding_opts & THINSTENC_DereferenceFirstOperand)
        {
            cmd.Op1.type   = o_phrase;
            cmd.Op1.phrase = reg1;
            signed char displacement = ua_next_byte();
            if(displacement != 0)
            {
                cmd.Op1.type   = o_displ;
                cmd.Op1.addr   = displacement; // displacement
            }
        }
        
        // we have a straight-up register
        else
        {
            cmd.Op1.type = o_reg;
            cmd.Op1.reg  = reg1;
        }
    }
    
    // we have an immediate
    else
    {
        unsigned long second_value = 0;
        
        if(width == 1)
            second_value = ua_next_byte();
        else
        if(width == 2)
            second_value = ua_next_word();
        else
        if(width == 4)
            second_value = ua_next_long();

        cmd.Op1.type  = o_imm;
        cmd.Op1.value = second_value;
        length  += (width - 1);
    }
    return length;
}

unsigned long ana_th_13_to_1b(unsigned long code)
{
    unsigned long width          = 1; // bitness of registers, immediates
    unsigned char length         = 2;
    unsigned char encoding_opts  = ua_next_byte();

    switch(code)
    {
    case 19:
        cmd.itype = TH_jmp;
        break;
    case 20:
        cmd.itype = TH_jz;
        break;
    case 21:
        cmd.itype = TH_jnz;
        break;
    case 22:
        cmd.itype = TH_jl;
        break;
    case 23:
        cmd.itype = TH_jle;
        break;
    case 24:
        cmd.itype = TH_jg;
        break;
    case 25:
        cmd.itype = TH_jge;
        break;
    case 26:
        cmd.itype = TH_vmcall;
        break;
    case 27:
        cmd.itype = TH_x86call;
        break;
    }

    if(encoding_opts & THINSTENC_LargeWidthModifier)
        width += 2;

    if( !(encoding_opts & THINSTENC_SmallWidthModifier) )
        width += 1;

    TH_regs       reg1              = rEND;
    bool bFirstOperandIsRegister    = 0;
    bool bFirstOperandIsHeapKey     = 0;
    unsigned long extended_heap_key = 0;
    unsigned long next_eip          = 0;

    if(encoding_opts & THINSTENC_FirstOperandIsRegister)    
    {
        reg1 = TH_Regnum(width, encoding_opts & 0xf);
        bFirstOperandIsRegister = 1;
    }
    
    // we are dealing with an immediate value
    else
    {
        // we have a heap regions 
        if(encoding_opts & THINSTENC_HeapMemoryKeyFollows)
        {
            // extended_heap_key contains the key to jump to (with 
            // possible displacement)
            bFirstOperandIsHeapKey = 1;

            // heap region is keyed with a single byte
            if(encoding_opts & THINSTENC_OneByteDisplacement)
            {
                extended_heap_key = ua_next_byte();
                length += 1;
            }
            // we have a 32-bit immediate heap memory key
            else
            {
                extended_heap_key = ua_next_long();
                length += 4;
            }
        } // end dealing with heap regions
        
        else
        {
            // short jumps (single byte displacements)
            if(encoding_opts & THINSTENC_OneByteDisplacement)
            {
                next_eip = (signed char)ua_next_byte() + cmd.ea;
                length += 1;
            }
            // we have a 32-bit immediate address
            else
            {
                next_eip = ua_next_long();
                length += 4;
            }
            msg("%lx:  ana():  calculated eip @ %lx\n", cmd.ea, next_eip);
        }
    }
    
    if(encoding_opts & THINSTENC_DereferenceOperand)
    {
        if(bFirstOperandIsRegister)
        {
            cmd.Op1.type   = o_phrase;
            cmd.Op1.phrase = reg1;
        }
        else
        if(bFirstOperandIsHeapKey)
        {
            // the VM allows you to pull up the contents of a heap region
            // and dereference it in the same instruction.
            // gonna need specflags for this one.

            // This scenario never materialized, so forget it.
        }
        else
        {
        }
    }
    else
    {
        if(bFirstOperandIsRegister)
        {
            cmd.Op1.type = o_reg;
            cmd.Op1.reg  = reg1;
        }
        else
        if(bFirstOperandIsHeapKey)
        {
            cmd.Op1.type = o_vmmem;
            cmd.Op1.addr = extended_heap_key;
        }
        else
        {
            cmd.Op1.type = o_near;
            cmd.Op1.addr = next_eip;
        }
    }
    return length;
}

unsigned long ana_th_1c_to_1f(unsigned long code)
{
    unsigned long length = 0;
    
    switch(code)
    {
    case 0x1c:
        cmd.itype     = TH_valloc;
        cmd.Op1.value = ua_next_long();
        cmd.Op1.dtyp  = dt_dword;
        cmd.Op1.type  = o_imm;
        cmd.Op2.value = ua_next_long();
        cmd.Op2.dtyp  = dt_dword;
        cmd.Op2.type  = o_imm;
        length       += 9;
        break;
    case 0x1d:
        cmd.itype     = TH_vfree;
        cmd.Op1.value = ua_next_long();
        cmd.Op1.dtyp  = dt_dword;
        cmd.Op1.type  = o_imm;
        length += 5;
        break;
    case 0x1e:
        cmd.itype     = TH_halloc;
        cmd.Op1.value = ua_next_long();
        cmd.Op1.dtyp  = dt_dword;
        cmd.Op1.type  = o_imm;
        cmd.Op2.value = ua_next_long();
        cmd.Op2.dtyp  = dt_dword;
        cmd.Op2.type  = o_imm;
        length       += 9;
        break;        
    case 0x1f:        
        cmd.itype     = TH_hfree;
        cmd.Op1.value = ua_next_long();
        cmd.Op1.dtyp  = dt_dword;
        cmd.Op1.type  = o_imm;
        length += 5;
        break;
    }
    return length;
}

unsigned long ana_th_20_to_24(unsigned long code)
{
    unsigned long length = 0;
    
    if(code != 0x24)
    {
        cmd.Op1.type = o_reg;
        cmd.Op1.reg  = TH_Regnum(4, (ua_next_byte() & 0xf0) >> 4);
        cmd.Op1.dtyp = dt_dword;
        length += 2;
    }
    switch(code)
    {
    case 0x20:
        cmd.itype     = TH_geteip;
        break;
    case 0x21:
        cmd.itype     = TH_getesp;
        break;        
    case 0x22:
        cmd.itype     = TH_getefl;
        break;
    case 0x23:        
        cmd.itype     = TH_getmem;
        cmd.Op2.addr  = ua_next_long();
        cmd.Op2.dtyp  = dt_dword;
        cmd.Op2.type  = o_vmmem;
        length += 4;
        break;
    case 0x24:
        cmd.itype     = TH_retd;
        cmd.Op1.value = ua_next_word();
        cmd.Op1.dtyp  = dt_word;
        cmd.Op1.type  = o_imm;
        length += 3;
        break;
    }
    return length;
}

int ana( void )
{
    cmd.Op1.type = o_void;
    cmd.Op2.type = o_void;
  
    uint code = ua_next_byte();
    if( (code >= 0) && (code <= 0xD) )
    {
        cmd.size  = ana_th_0_to_0d(code);
    }
    else
    if( (code >= 0xE) && (code <= 0x12) )
    {
        cmd.size  = ana_th_0e_to_12(code);
    }
    else
    if( (code >= 0x13) && (code <= 0x1B) )
    {
        cmd.size  = ana_th_13_to_1b(code);
    }
    else
    if( (code >= 0x1C) && (code <= 0x1F) )
    {
        cmd.size  = ana_th_1c_to_1f(code);
    }
    else
    if( (code >= 0x20) && (code <= 0x24) )
    {
        cmd.size  = ana_th_20_to_24(code);
    }
    else
    if( code == 0xff )
    {
        cmd.itype = TH_stop;
        cmd.size  = 1;
    }
    else
    {
        cmd.itype = TH_null;
        cmd.size  = 0;
        // not a type
    }
    return cmd.size;
}
