/*
 *  Interactive disassembler (IDA).
 *  TheHyper VM Processor Module.
 *
 *  (c) 2006 Rolf Rolles
 */

#include <ida.hpp>
#include <idp.hpp>
#include "ins.hpp"

instruc_t Instructions[] = {
  { "",                     0                       },  // Unknown Operation
  
  // GROUP 1:  Two-Operand Arithmetic Instructions
  { "mov" ,   CF_USE2 | CF_USE1 | CF_CHG1           },  // Move
  { "add" ,   CF_USE2 | CF_USE1 | CF_CHG1           },  // Add
  { "sub" ,   CF_USE2 | CF_USE1 | CF_CHG1           },  // Subtract
  { "xor" ,   CF_USE2 | CF_USE1 | CF_CHG1           },  // Exclusive Bitwise OR
  { "and" ,   CF_USE2 | CF_USE1 | CF_CHG1           },  // Bitwise And
  { "or"  ,   CF_USE2 | CF_USE1 | CF_CHG1           },  // Bitwise OR
  { "imul",   CF_USE2 | CF_USE1 | CF_CHG1           },  // Integer Multiply
  { "idiv",   CF_USE2 | CF_USE1 | CF_CHG1           },  // Integer Divide
  { "imod",   CF_USE2 | CF_USE1 | CF_CHG1           },  // Integer Modulus
  { "ror" ,   CF_USE2 | CF_USE1 | CF_CHG1           },  // Bitwise Rotate Right
  { "rol" ,   CF_USE2 | CF_USE1 | CF_CHG1           },  // Bitwise Rotate Left
  { "shr" ,   CF_USE2 | CF_USE1 | CF_CHG1           },  // Bitwise Shift Right
  { "shl" ,   CF_USE2 | CF_USE1 | CF_CHG1           },  // Bitwise Shift Left
  { "cmp" ,   CF_USE2 | CF_USE1                     },  // Comparison Operator
  
  // GROUP 2:  One-Operand Arithmetic And General Instruction
  { "push",             CF_USE1                     },  // Push To Stack
  { "pop" ,             CF_USE1 | CF_CHG1           },  // Pop From Stack
  { "inc" ,             CF_USE1 | CF_CHG1           },  // Increment Operator
  { "dec" ,             CF_USE1 | CF_CHG1           },  // Decrement Operator
  { "not" ,             CF_USE1 | CF_CHG1           },  // Bitwise Not Operator
  
  // GROUP 3:  Conditional Instructions
  { "jmp" ,             CF_USE1           | CF_STOP },  // Unconditional Branch
  { "jz"  ,             CF_USE1                     },  // Conditional (Zero         ) Branch Positive
  { "jnz" ,             CF_USE1                     },  // Conditional (Zero         ) Branch Negative
  { "jl"  ,             CF_USE1                     },  // Conditional (Lesser       ) Branch Positive
  { "jle" ,             CF_USE1                     },  // Conditional (Lesser-Equal ) Branch Negative
  { "jg"  ,             CF_USE1                     },  // Conditional (Greater      ) Branch Positive
  { "jge" ,             CF_USE1                     },  // Conditional (Greater-Equal) Branch Positive
  { "vmcall",           CF_USE1           | CF_CALL },  // Do call inside of VM
  { "x86call",          CF_USE1                     },  // Do call inside of host process' address space
  
  // GROUP 4:  Memory-Related Instructions
  { "valloc", CF_USE2 | CF_USE1                     },  // Alloc virtual memory via VirtualAlloc
  { "vfree" ,           CF_USE1                     },  // Free  virtual memory via VirtualFree
  { "halloc", CF_USE2 | CF_USE1                     },  // Alloc heap memory via HeapAlloc
  { "hfree" ,           CF_USE1                     },  // Free  heap memory via HeapFree
                                                    
  // GROUP 5:  Miscellaneous Instructions           
  { "geteip" ,          CF_USE1 | CF_CHG1           },  // Get current EIP
  { "getesp" ,          CF_USE1 | CF_CHG1           },  // Get current ESP
  { "getefl" ,          CF_USE1 | CF_CHG1           },  // Get EFLAGS
  { "getmem" ,          CF_USE1 | CF_CHG1           },  // Get memory block by key
  { "ret"    ,          CF_USE1           | CF_STOP },  // Ret [STACKPOP VALUE]
  { "stop"   ,                              CF_STOP }   // No more instructions

};
